@@ -25,22 +25,16 @@
 
 "Export to CSV Spreadsheet."
 
-
-from __future__ import unicode_literals
 #-------------------------------------------------------------------------
 #
 # Standard Python Modules
 #
 #-------------------------------------------------------------------------
 import os
-import sys
 from gramps.gen.const import GRAMPS_LOCALE as glocale
 _ = glocale.translation.sgettext
 import csv
-if sys.version_info[0] < 3:
-    from cStringIO import StringIO
-else:
-    from io import StringIO
+from io import StringIO
 import codecs
 
 #------------------------------------------------------------------------
@@ -49,12 +43,12 @@ import codecs
 #
 #------------------------------------------------------------------------
 import logging
-import collections
+from collections import abc
 LOG = logging.getLogger(".ExportCSV")
 
 #-------------------------------------------------------------------------
 #
-# GRAMPS modules
+# Gramps modules
 #
 #-------------------------------------------------------------------------
 from gramps.gen.lib import EventType, Person
@@ -64,6 +58,7 @@ from gramps.gen.utils.string import gender as gender_map
 from gramps.gen.datehandler import get_date
 from gramps.gen.display.place import displayer as _pd
 from gramps.gui.glade import Glade
+from gramps.gen.constfunc import win
 
 #-------------------------------------------------------------------------
 #
@@ -81,7 +76,7 @@ def exportData(database, filename, user, option_box=None):
 #-------------------------------------------------------------------------
 def sortable_string_representation(text):
     numeric = ""
-    alpha   = ""
+    alpha = ""
     for s in text:
         if s.isdigit():
             numeric += s
@@ -108,79 +103,6 @@ def get_primary_source_title(db, obj):
             return source.get_title()
     return ""
 
-#-------------------------------------------------------------------------
-#
-# Encoding support for CSV, from http://docs.python.org/lib/csv-examples.html
-#
-#-------------------------------------------------------------------------
-class UTF8Recoder(object):
-    """Iterator that reads an encoded stream and reencodes the input to UTF-8."""
-    def __init__(self, f, encoding):
-        self.reader = codecs.getreader(encoding)(f)
-
-    def __iter__(self):
-        return self
-
-    def __next__(self):
-        return self.reader.next().encode("utf-8")
-
-class UnicodeReader(object):
-    """
-    A CSV reader which will iterate over lines in the CSV file "f", which is 
-    encoded in the given encoding.
-    
-    """
-
-    def __init__(self, f, encoding="utf-8", **kwds):
-        f = UTF8Recoder(f, encoding)
-        self.reader = csv.reader(f, **kwds)
-
-    def __next__(self):
-        row = next(self.reader)
-        return [str(s, "utf-8") for s in row]
-
-    def __iter__(self):
-        return self
-
-class UnicodeWriter(object):
-    """
-    A CSV writer which will write rows to CSV file "f", which is encoded in 
-    the given encoding.
-    
-    """
-
-    def __init__(self, f, encoding="utf-8", **kwds):
-        # Redirect output to a queue
-        self.queue = StringIO()
-        self.writer = csv.writer(self.queue, **kwds)
-        self.stream = f
-        self.encoder = codecs.getencoder(encoding)
-
-    def writerow(self, row):
-        if sys.version_info[0] < 3:
-            self.writer.writerow([s.encode('utf-8') for s in row])
-            # Fetch UTF-8 output from the queue ...
-            data = self.queue.getvalue()
-            data = data.decode('utf-8')
-        else:
-            self.writer.writerow(row)
-            data = self.queue.getvalue()
-            #in python3, StringIO self.queue returns unicode!
-        #data now contains the csv data in unicode
-        # ... and reencode it into the target encoding
-        data, length = self.encoder(data)
-        # write to the target stream
-        self.stream.write(data)
-        # empty queue, go to start position, then truncate
-        self.queue.seek(0)
-        self.queue.truncate(0)
-
-    def writerows(self, rows):
-        list(map(self.writerow, rows))
-
-    def close(self):
-        self.stream.close()
-
 #-------------------------------------------------------------------------
 #
 # CSVWriter Options
@@ -189,18 +111,22 @@ class UnicodeWriter(object):
 class CSVWriterOptionBox(WriterOptionBox):
     """
     Create a VBox with the option widgets and define methods to retrieve
-    the options. 
-    
+    the options.
+
     """
-    def __init__(self, person, dbstate, uistate):
-        WriterOptionBox.__init__(self, person, dbstate, uistate)
+    def __init__(self, person, dbstate, uistate, track=[], window=None):
+        WriterOptionBox.__init__(self, person, dbstate, uistate, track=track,
+                                 window=window)
+        ## TODO: add place filter selection
         self.include_individuals = 1
         self.include_marriages = 1
         self.include_children = 1
+        self.include_places = 1
         self.translate_headers = 1
         self.include_individuals_check = None
         self.include_marriages_check = None
         self.include_children_check = None
+        self.include_places_check = None
         self.translate_headers_check = None
 
     def get_option_box(self):
@@ -210,16 +136,19 @@ class CSVWriterOptionBox(WriterOptionBox):
         self.include_individuals_check = Gtk.CheckButton(label=_("Include people"))
         self.include_marriages_check = Gtk.CheckButton(label=_("Include marriages"))
         self.include_children_check = Gtk.CheckButton(label=_("Include children"))
+        self.include_places_check = Gtk.CheckButton(label=_("Include places"))
         self.translate_headers_check = Gtk.CheckButton(label=_("Translate headers"))
 
-        self.include_individuals_check.set_active(1) 
-        self.include_marriages_check.set_active(1) 
-        self.include_children_check.set_active(1) 
-        self.translate_headers_check.set_active(1) 
+        self.include_individuals_check.set_active(1)
+        self.include_marriages_check.set_active(1)
+        self.include_children_check.set_active(1)
+        self.include_places_check.set_active(1)
+        self.translate_headers_check.set_active(1)
 
         option_box.pack_start(self.include_individuals_check, False, True, 0)
         option_box.pack_start(self.include_marriages_check, False, True, 0)
         option_box.pack_start(self.include_children_check, False, True, 0)
+        option_box.pack_start(self.include_places_check, False, True, 0)
         option_box.pack_start(self.translate_headers_check, False, True, 0)
 
         return option_box
@@ -230,6 +159,7 @@ class CSVWriterOptionBox(WriterOptionBox):
             self.include_individuals = self.include_individuals_check.get_active()
             self.include_marriages = self.include_marriages_check.get_active()
             self.include_children = self.include_children_check.get_active()
+            self.include_places = self.include_places_check.get_active()
             self.translate_headers = self.translate_headers_check.get_active()
 
 #-------------------------------------------------------------------------
@@ -237,28 +167,30 @@ class CSVWriterOptionBox(WriterOptionBox):
 # CSVWriter class
 #
 #-------------------------------------------------------------------------
-class CSVWriter(object):
+class CSVWriter:
     def __init__(self, database, filename, user, option_box=None):
         self.db = database
         self.option_box = option_box
         self.filename = filename
         self.user = user
-        if isinstance(self.user.callback, collections.Callable): # callback is really callable
+        if isinstance(self.user.callback, abc.Callable):  # is really callable
             self.update = self.update_real
         else:
             self.update = self.update_empty
 
         self.plist = {}
         self.flist = {}
-        
+        self.place_list = {}
+
         self.persons_details_done = []
         self.persons_notes_done = []
         self.person_ids = {}
-        
+
         if not option_box:
             self.include_individuals = 1
             self.include_marriages = 1
             self.include_children = 1
+            self.include_places = 1
             self.translate_headers = 1
         else:
             self.option_box.parse_options()
@@ -267,9 +199,24 @@ class CSVWriter(object):
             self.include_individuals = self.option_box.include_individuals
             self.include_marriages = self.option_box.include_marriages
             self.include_children = self.option_box.include_children
+            self.include_places = self.option_box.include_places
             self.translate_headers = self.option_box.translate_headers
-            
+
         self.plist = [x for x in self.db.iter_person_handles()]
+
+        # make place list so that dependencies are first:
+        self.place_list = []
+        place_list = sorted([x for x in self.db.iter_place_handles()])
+        while place_list:
+            handle = place_list[0]
+            place = self.db.get_place_from_handle(handle)
+            if place:
+                if all([(x.ref in self.place_list) for x in place.placeref_list]):
+                    self.place_list.append(place_list.pop(0))
+                else: # put at the back of the line:
+                    place_list.append(place_list.pop(0))
+            else:
+                place_list.pop(0)
         # get the families for which these people are spouses:
         self.flist = {}
         for key in self.plist:
@@ -286,7 +233,7 @@ class CSVWriter(object):
                         child_handle = child_ref.ref
                         if child_handle in self.plist:
                             self.flist[family_handle] = 1
-                        
+
     def update_empty(self):
         pass
 
@@ -306,9 +253,10 @@ class CSVWriter(object):
     def export_data(self):
         self.dirname = os.path.dirname (self.filename)
         try:
-            self.g = open(self.filename,"w")
-            self.fp = open(self.filename, "wb")
-            self.g = UnicodeWriter(self.fp)
+            self.fp = open(self.filename, "w",
+                           encoding='utf_8_sig' if win() else 'utf_8',
+                           newline='')
+            self.g = csv.writer(self.fp)
         except IOError as msg:
             msg2 = _("Could not create %s") % self.filename
             self.user.notify_error(msg2,str(msg))
@@ -326,9 +274,50 @@ class CSVWriter(object):
             self.total += len(self.flist)
         if self.include_children:
             self.total += len(self.flist)
-        ######################## 
+        if self.include_places:
+            self.total += len(self.place_list)
+        ########################
         LOG.debug("Possible people to export: %s", len(self.plist))
         LOG.debug("Possible families to export: %s", len(self.flist))
+        LOG.debug("Possible places to export: %s", len(self.place_list))
+        ###########################
+        if self.include_places:
+            if self.translate_headers:
+                self.write_csv(_("Place"), _("Title"), _("Name"),
+                               _("Type"), _("Latitude"), _("Longitude"),
+                               _("Code"), _("Enclosed_by"), _("Date"))
+            else:
+                self.write_csv("Place", "Title", "Name",
+                               "Type", "Latitude", "Longitude",
+                               "Code", "Enclosed_by", "Date")
+            for key in self.place_list:
+                place = self.db.get_place_from_handle(key)
+                if place:
+                    place_id = place.gramps_id
+                    place_title = place.title
+                    place_name = place.name.value
+                    place_type = str(place.place_type)
+                    place_latitude = place.lat
+                    place_longitude = place.long
+                    place_code = place.code
+                    if place.placeref_list:
+                        for placeref in place.placeref_list:
+                            placeref_obj = self.db.get_place_from_handle(placeref.ref)
+                            placeref_date = ""
+                            if not placeref.date.is_empty():
+                                placeref_date = placeref.date
+                            placeref_id = ""
+                            if placeref_obj:
+                                placeref_id = "[%s]" % placeref_obj.gramps_id
+                            self.write_csv("[%s]" % place_id, place_title, place_name, place_type,
+                                           place_latitude, place_longitude, place_code, placeref_id,
+                                           placeref_date)
+                    else:
+                        self.write_csv("[%s]" % place_id, place_title, place_name, place_type,
+                                       place_latitude, place_longitude, place_code, "",
+                                       "")
+                self.update()
+            self.writeln()
         ########################### sort:
         sortorder = []
         dropped_surnames = set()
@@ -352,7 +341,7 @@ class CSVWriter(object):
                         "{count} dropped").format(
                             count=len(dropped_surnames)) )
             LOG.debug(
-                    "Dropped surnames: " + 
+                    "Dropped surnames: " +
                     ', '.join([("%s %s %s" % (surname.get_prefix(),
                     surname.get_surname(), surname.get_connector())).strip()
                     for surname in dropped_surnames]))
@@ -362,22 +351,22 @@ class CSVWriter(object):
         if self.include_individuals:
             if self.translate_headers:
                 self.write_csv(
-                    _("Person"), _("Surname"), _("Given"), 
-                    _("Call"), _("Suffix"), _("Prefix"), 
-                    _("Person|Title"), _("Gender"), 
+                    _("Person"), _("Surname"), _("Given"),
+                    _("Call"), _("Suffix"), _("Prefix"),
+                    _("Person|Title"), _("Gender"),
                     _("Birth date"), _("Birth place"), _("Birth source"),
                     _("Baptism date"), _("Baptism place"), _("Baptism source"),
-                    _("Death date"), _("Death place"), _("Death source"), 
+                    _("Death date"), _("Death place"), _("Death source"),
                     _("Burial date"), _("Burial place"), _("Burial source"),
                     _("Note"))
             else:
                 self.write_csv(
-                    "Person", "Surname", "Given", 
-                    "Call", "Suffix", "Prefix", 
-                    "Title", "Gender", 
+                    "Person", "Surname", "Given",
+                    "Call", "Suffix", "Prefix",
+                    "Title", "Gender",
                     "Birth date", "Birth place", "Birth source",
                     "Baptism date", "Baptism place", "Baptism source",
-                    "Death date", "Death place", "Death source", 
+                    "Death date", "Death place", "Death source",
                     "Burial date", "Burial place", "Burial source",
                     "Note")
             for key in plist:
@@ -412,9 +401,7 @@ class CSVWriter(object):
                         birth = self.db.get_event_from_handle(birth_ref.ref)
                         if birth:
                             birthdate = self.format_date( birth)
-                            place_handle = birth.get_place_handle()
-                            if place_handle:
-                                birthplace = _pd.display_event(self.db, birth)
+                            birthplace = self.format_place(birth)
                             birthsource = get_primary_source_title(self.db, birth)
                     # Baptism:
                     baptismdate = ""
@@ -426,9 +413,7 @@ class CSVWriter(object):
                         baptism = self.db.get_event_from_handle(baptism_ref.ref)
                         if baptism:
                             baptismdate = self.format_date( baptism)
-                            place_handle = baptism.get_place_handle()
-                            if place_handle:
-                                baptismplace = _pd.display_event(self.db, baptism)
+                            baptismplace = self.format_place(baptism)
                             baptismsource = get_primary_source_title(self.db, baptism)
                     # Death:
                     deathdate = ""
@@ -439,9 +424,7 @@ class CSVWriter(object):
                         death = self.db.get_event_from_handle(death_ref.ref)
                         if death:
                             deathdate = self.format_date( death)
-                            place_handle = death.get_place_handle()
-                            if place_handle:
-                                deathplace = _pd.display_event(self.db, death)
+                            deathplace = self.format_place(death)
                             deathsource = get_primary_source_title(self.db, death)
                     # Burial:
                     burialdate = ""
@@ -453,9 +436,7 @@ class CSVWriter(object):
                         burial = self.db.get_event_from_handle(burial_ref.ref)
                         if burial:
                             burialdate = self.format_date( burial)
-                            place_handle = burial.get_place_handle()
-                            if place_handle:
-                                burialplace = _pd.display_event(self.db, burial)
+                            burialplace = self.format_place(burial)
                             burialsource = get_primary_source_title(self.db, burial)
                     # Write it out:
                     self.write_csv(grampsid_ref, surname, first_name, callname,
@@ -478,13 +459,13 @@ class CSVWriter(object):
                     )
         sortorder.sort() # will sort on tuples
         flist = [data[1] for data in sortorder]
-        ########################### 
+        ###########################
         if self.include_marriages:
             if self.translate_headers:
-                self.write_csv(_("Marriage"), _("Husband"), _("Wife"), 
+                self.write_csv(_("Marriage"), _("Husband"), _("Wife"),
                                _("Date"), _("Place"), _("Source"), _("Note"))
             else:
-                self.write_csv("Marriage", "Husband", "Wife", 
+                self.write_csv("Marriage", "Husband", "Wife",
                                "Date", "Place", "Source", "Note")
             for key in flist:
                 family = self.db.get_family_from_handle(key)
@@ -513,10 +494,8 @@ class CSVWriter(object):
                         event = self.db.get_event_from_handle(event_ref.ref)
                         if event.get_type() == EventType.MARRIAGE:
                             mdate = self.format_date( event)
-                            place_handle = event.get_place_handle()
-                            if place_handle:
-                                mplace = _pd.display_event(self.db, event)
-                                source = get_primary_source_title(self.db, event)
+                            mplace = self.format_place(event)
+                            source = get_primary_source_title(self.db, event)
                     note = ''
                     self.write_csv(marriage_id, father_id, mother_id, mdate,
                                    mplace, source, note)
@@ -543,8 +522,23 @@ class CSVWriter(object):
                         self.write_csv(family_id, grampsid_ref)
                 self.update()
             self.writeln()
-        self.g.close()
-        return True 
-    
+        self.fp.close()
+        return True
+
     def format_date(self, date):
         return get_date(date)
+
+    def format_place(self, event):
+        """
+        If places are included in the export return a link, else return a
+        formatted place for the given event.
+        """
+        if self.include_places:
+            place_handle = event.get_place_handle()
+            if place_handle:
+                place = self.db.get_place_from_handle(place_handle)
+                if place:
+                    return "[%s]" % place.get_gramps_id()
+            return ""
+        else:
+            return _pd.display_event(self.db, event)
